#!/usr/bin/env python
# Filename: 	clean_up_annotations.py
# Author: 	Alessandro Gerada
# Date: 	2023-03-17
# Copyright: 	Alessandro Gerada 2023
# Email: 	alessandro.gerada@liverpool.ac.uk

"""Script that cleans up duplicate images in annotation folders. Use clean_up_annotations.py -h for help."""

import argparse
import cv2  # pylint: disable=import-error
from aigarmic._img_utils import get_image_paths, Deleter, in_list


def clean_up_annotations_parser():
    parser = argparse.ArgumentParser("""
        Clean up duplicate images in annotation folders
        """)
    parser.add_argument('input_dir', type=str,
                        help="Input directory - can contain subdirectory of images which will be processed separately")
    parser.add_argument('-q', '--quiet', action='store_true',
                        help="Suppress file deletion warnings (CAUTION)")
    return parser


def main():
    parser = clean_up_annotations_parser()
    args = parser.parse_args()

    images_paths = get_image_paths(args.input_dir)
    
    deleter = Deleter(confirm=False if args.quiet else True)

    _images_list = []
    if isinstance(images_paths, list):
        for i in images_paths: 
            _image = cv2.imread(i)
            if not in_list(_image, _images_list): 
                _images_list.append(_image)
            else: 
                deleter.delete_file(i)

    if isinstance(images_paths, dict):
        for _, v in images_paths.items():
            _images_list = []
            for i in v: 
                _image = cv2.imread(i)
                if not in_list(_image, _images_list): 
                    _images_list.append(_image)
                else: 
                    deleter.delete_file(i)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# Filename: 	main.py
# Author: 	Alessandro Gerada
# Date: 	2023-01-27
# Copyright: 	Alessandro Gerada 2023
# Email: 	alessandro.gerada@liverpool.ac.uk

"""Script to predict MICs from agar dilution images. Use main.py -h for help."""

import pathlib
from aigarmic.process_plate_image import split_by_grid
from aigarmic.plate import plate_set_from_dir
import argparse
from aigarmic.file_handlers import get_concentration_from_path, get_paths_from_directory
import csv
from aigarmic.model import SoftmaxModel, BinaryModel, BinaryNestedModel
import sys
import cv2  # pylint: disable=import-error


def main_parser():
    parser = argparse.ArgumentParser(description="Main script to interpret agar dilution MICs",
                                     formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument('directory', type=str, help="""
            Directory containing images to process, arranged in sub-folders by antibiotic name, e.g.,: \n
            \t directory/ \n
            \t \t antibiotic1_name/ \n
            \t \t \t 0.jpg \n
            \t \t \t 0.125.jpg \n
            \t \t \t 0.25.jpg \n
            """)
    parser.add_argument("-m", "--model", type=str, nargs="*",
                        help="Specify one or more directories containing tensorflow model/s for image classification")
    parser.add_argument("-t", "--type_model", type=str, default="binary",
                        help="Type of keras model, e.g., binary, softmax [default] = binary")
    parser.add_argument("-o", "--output_file", type=str,
                        help="Specify output file for csv report (will be overwritten)")
    parser.add_argument("-v", "--validation_threshold", type=float, default=0.0,
                        help="Image annotation accuracies below this value will be validated for manual confirmation."
                             "0.0 checks no images [default]; 1.0 checks all images.")
    parser.add_argument("-c", "--check_contours", action="store_true", help="Check contours visually")
    parser.add_argument("-n", "--negative_codes", type=str,
                        help="Comma-separated list of no growth class codes for model, e.g., 0,1 (default=0)")
    parser.add_argument("-s", "--softmax_classes", type=int,
                        help="Number of softmax classes for softmax classes that model predicts."
                             "Required if -t = softmax")
    parser.add_argument("-d", "--dimensions", type=int, nargs=2, default=[160, 160],
                        help="X and Y dimensions of images for model training [default = 160 160]")
    return parser


def main():
    parser = main_parser()
    args = parser.parse_args()

    model_image_x = args.dimensions[0]
    model_image_y = args.dimensions[1]
    supported_model_types = ['softmax', 'binary']

    plate_images_paths = get_paths_from_directory(args.directory)

    if args.check_contours:
        cv2.startWindowThread()
        for abx, paths in plate_images_paths.items():
            for path in paths:
                _image = cv2.imread(path)
                try:
                    split_by_grid(_image,
                                  visualise_contours=True,
                                  plate_name=abx + '_' + str(get_concentration_from_path(path)))
                except ValueError as e:
                    raise ValueError from e

        pos_replies = ['y', 'yes', 'ye']
        neg_replies = ['n', 'no']
        cv2.waitKey(1)
        cv2.destroyAllWindows()
        cv2.waitKey(1)
        while True:
            input_key = input("""Completed contour checks. Would you like to continue with annotation? [Y / N]
                              Please only proceed if all images have correctly identified 96 boxes!
                              """)
            input_key = input_key.lower()
            if input_key in neg_replies:
                sys.exit()
            elif input_key in pos_replies:
                print("Continuing with annotation..")
                break
            else:
                print("Unable to recognise input, please try again..")
                continue

    if args.type_model not in supported_model_types:
        sys.exit(f"Model type specified is not supported, please use one of {supported_model_types}")

    if args.type_model == 'softmax' and len(args.model) != 1:
        sys.exit(
            """
            Softmax model can only run with one keras model
            """
        )

    if args.type_model == 'softmax' and not args.softmax_classes:
        sys.exit(
            """
            Softmax model requires the number of classes to be specified
            """
        )

    if args.type_model == 'softmax':
        # Since args.model is a list, un-list
        [path_to_model] = args.model
        model = SoftmaxModel(path_to_model, trained_x=model_image_x, trained_y=model_image_y,
                             key=[str(x) for x in range(args.softmax_classes)])

    elif args.type_model == 'binary' and len(args.model) == 2:
        class_names_first_line = ['No growth', 'Growth']
        class_names_second_line = ['Poor growth', 'Good growth']
        first_line_model = BinaryModel(args.model[0], key=class_names_first_line, trained_x=model_image_x,
                                       trained_y=model_image_y)
        second_line_model = BinaryModel(args.model[1], key=class_names_second_line, trained_x=model_image_x,
                                        trained_y=model_image_y)
        model = BinaryNestedModel(first_line_model, second_line_model, first_model_accuracy_acceptance=0.6,
                                  suppress_first_model_accuracy_check=True)
    elif args.type_model == 'binary' and len(args.model) == 1:
        class_names = ['No growth', 'Growth']
        model = BinaryModel(args.model[0], key=class_names,
                            trained_x=model_image_x, trained_y=model_image_y)
    else:
        sys.exit(f"Model type specified is not supported, please use one of {supported_model_types}")

    abx_superset = {}
    parent_path = pathlib.Path(args.directory)
    for abx, paths in plate_images_paths.items():
        _plate_set = plate_set_from_dir(path=parent_path / abx,
                                        drug=abx,
                                        model=model)
        abx_superset[abx] = _plate_set

    for abx, plate_set in abx_superset.items():
        plate_set.review_poor_images(threshold=args.validation_threshold)
        if args.negative_codes:
            ng_codes = [int(x) for x in args.negative_codes.split(",")]
            plate_set.calculate_mic(no_growth_key_items=tuple(ng_codes))
        else:
            print("Warning: negative growth codes not specified, assuming any growth is positive."
                  "Please use -n to specify negative codes")
            plate_set.calculate_mic(no_growth_key_items=tuple([0]))
        plate_set.generate_qc()

    if args.output_file:
        output_data = []
        for plate_set in abx_superset.values():
            output_data = output_data + plate_set.get_csv_data()
        with open(args.output_file, 'w', encoding="utf-8-sig") as csvfile:
            writer = csv.DictWriter(csvfile, output_data[0].keys())
            writer.writeheader()
            writer.writerows(output_data)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# Filename: 	manual_annotator.py
# Author: 	Alessandro Gerada
# Date: 	2023-01-28
# Copyright: 	Alessandro Gerada 2023
# Email: 	alessandro.gerada@liverpool.ac.uk

"""
This script shows random images from 96-well plates, and waits for user input to 
annotate the image. Use the following key: 
0 = no growth
1 = faint growth or colony
2 = normal growth
The output is stored in annotations/
"""

from aigarmic.file_handlers import get_concentration_from_path, get_paths_from_directory
from aigarmic.plate import Plate
import os
import cv2  # pylint: disable=import-error
import argparse
from random import choice
from datetime import datetime


def manual_annotator_parser():
    parser = argparse.ArgumentParser(description="Manually annotate plate images")
    parser.add_argument('input_directory', type=str,
                        help="Directory containing plate images")
    parser.add_argument('-o', '--output_directory', type=str, default='annotations/',
                        help='Path to store annotation output files')
    return parser


def main(): 
    parser = manual_annotator_parser()
    args = parser.parse_args()
    codes = {}
    for ascii_code, class_code in zip(range(48, 58), range(0, 10)):
        codes[ascii_code] = class_code

    stop_codes = {27: "esc"}
    codes.update(stop_codes)

    input_dir = args.input_directory
    plate_image_paths = get_paths_from_directory(input_dir)
    plates = []
    for abx, paths in plate_image_paths.items(): 
        for path in paths: 
            try: 
                concentration = get_concentration_from_path(path)
                plates.append(Plate(abx, concentration, path, visualise_contours=False,
                                    n_row=8, n_col=12))
            except FileNotFoundError as e:
                print(f"Error while trying to import {path}: ")
                print(e)

    print("Use the following key for input: ")
    print("0 = no growth")
    print("1 = faint growth or colony")
    print("2 = normal growth")
    print("Press esc to cancel at any time. ")
    print()

    n = 0
    while True: 
        try: 
            n = int(input("How many images do you want to annotate? "))
        except ValueError: 
            print("Unable to parse input, try again.. ")
            continue
        break

    output_images = []
    output_annotations = []
    output_image_codes = []

    for i in range(n): 
        x, code = choice(plates).get_colony_image()
        code = code + '_' + datetime.now().strftime('%Y-%m-%d_%H%M%S')
        cv2.imshow('image', x)
        print("Please enter classification for this image..")
        while True: 
            input_key = cv2.waitKey()
            if input_key not in codes: 
                print("Input not recognised, please try again..")
                continue
            else: 
                break
        if input_key in stop_codes:
            print("Exiting.")
            break
        else: 
            output_images.append(x)
            output_annotations.append(codes[input_key])
            output_image_codes.append(code)

    output_dir = args.output_directory
    if not os.path.exists(output_dir): 
        os.mkdir(output_dir)

    # make class folders
    for a in output_annotations: 
        if not os.path.exists(os.path.join(output_dir, str(a))): 
            os.mkdir(os.path.join(output_dir, str(a)))
    
    for i, a, c in zip(output_images, output_annotations, output_image_codes):
        cv2.imwrite(os.path.join(output_dir, str(a), c + '.jpg'), i)


if __name__ == "__main__": 
    main()



from aigarmic.plate import Plate, PlateSet, plate_set_from_dir
from aigarmic.model import Model, BinaryModel, BinaryNestedModel, SoftmaxModel
from aigarmic.file_handlers import (create_dataset_from_directory,
                                    predict_colony_images_from_directory,
                                    save_training_log,
                                    get_paths_from_directory,
                                    get_concentration_from_path)
from aigarmic._img_utils import (convert_cv2_to_keras,
                                 keras_image_to_cv2)
from aigarmic.process_plate_image import find_threshold_value, split_by_grid
from aigarmic.train import train_binary, train_softmax, visualise_training



# Filename:      _img_utils.py
# Author:        Alessandro Gerada
# Date:          01/10/2022
# Copyright:     Alessandro Gerada 2022
# Email:         alessandro.gerada@liverpool.ac.uk

"""
Image utility functions
"""
from os import remove
from typing import Union
import cv2  # pylint: disable=import-error
import numpy as np
import os
import tensorflow as tf


def convert_cv2_to_keras(image, size_x=160, size_y=160) -> np.ndarray:
    """
    Convert a cv2 image to a keras image

    :param image: Image loaded using cv2.imread
    :param size_x: Width to resize image to (pixels)
    :param size_y: Height to resize image to (pixels)
    :return: Image as a numpy array
    """
    # resize
    image = cv2.resize(image, (size_x, size_y))
    # convert from BGR to RGB
    image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
    image = image.reshape(1, size_x, size_y, 3)
    image = image.astype(np.float32)
    return image


def keras_image_to_cv2(image: tf.Tensor) -> np.ndarray:
    """
    Convert a keras image to a cv2 image

    :param image: Image as a tensor
    :return: Image as a numpy array
    """
    img = image.numpy().astype(np.uint8)
    img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
    return img


def get_image_paths(dir_path,
                    extensions: tuple[str, ...] = ('.jpg', '.JPG')) -> Union[list[str], dict[str, list[str]]]:
    """
    If there are no subdirectories in dir, returns a list of image paths
    If there are subdirectories, returns a dict of 'subdir_name': 'path'

    :param dir_path: Path to directory containing images
    :param extensions: Tuple of image extensions to search for
    :return: List of image paths or dict of 'subdir_name': 'path'
    """
    sub_dirs = [i for i in os.listdir(dir_path) if not i.startswith('.') and os.path.isdir(os.path.join(dir_path, i))]
    
    if not sub_dirs:
        return [os.path.join(dir_path, i) for i in os.listdir(dir_path) if i.endswith(tuple(extensions))]
    else: 
        output = {}
        for i in sub_dirs:
            _parent_path = os.path.join(dir_path, i)
            _temp_image_paths = os.listdir(_parent_path)
            _temp_image_paths = [i for i in _temp_image_paths if i.endswith(tuple(extensions))]
            _temp_image_paths = [os.path.join(_parent_path, i) for i in _temp_image_paths]
            output[i] = _temp_image_paths
        return output


class Deleter:
    def __init__(self, confirm=True):
        self.confirm = confirm
        self.converter = {
            "y": "yes",
            "ye": "yes",
            "yes": "yes",
            "n": "no",
            "no": "no",
            "a": "all",
            "al": "all",
            "all": "all"
        }

    def delete_file(self, file):
        if not self.confirm:
            print(f"Deleting duplicate file: {file}")
            remove(file)
        else:
            while True:
                user_in = input(f"[ Yes / No / All ] Delete duplicate file: {file}").lower()
                if user_in not in self.converter:
                    print("Input not recognised, please repeat")
                    continue
                if self.converter[user_in] == "yes":
                    print(f"Deleting file: {file}")
                    remove(file)
                    break
                if self.converter[user_in] == "no":
                    print(f"Skipping file: {file}")
                    break
                if self.converter[user_in] == "all":
                    print("Warnings suppressed")
                    print(f"Deleting file: {file}")
                    remove(file)
                    self.confirm = False
                    break
                else:
                    raise LookupError("Unable to interpret user loop")


def is_similar(image1, image2):
    """
    From: https://stackoverflow.com/a/23199159
    """
    return image1.shape == image2.shape and not (np.bitwise_xor(image1, image2).any())


def in_list(image, target_list):
    if not target_list:
        return False
    else:
        for i in target_list:
            if is_similar(i, image):
                return True
        return False


def rename_jpg(file_in, file_out):
    if not file_in or not file_out:
        return
    print(f"Renaming {file_in} to {file_out}")
    os.rename(file_in, file_out)



#!/usr/bin/env python
# Filename: 	train_modular.py
# Author: 	Alessandro Gerada
# Date: 	2023-08-11
# Copyright: 	Alessandro Gerada 2023
# Email: 	alessandro.gerada@liverpool.ac.uk

"""Modular training function that works with binary or softmax models. Use train_modular.py -h for help."""

from aigarmic.train import visualise_training, train_binary, train_softmax
from aigarmic.file_handlers import save_training_log
from aigarmic._nn_design import (model_design_spectrum_2024_binary_first_step,
                                 model_design_spectrum_2024_binary_second_step)
from aigarmic.file_handlers import predict_colony_images_from_directory
import os
import cv2  # pylint: disable=import-error
import argparse
import pathlib
import warnings


def train_modular_parser():
    parser = argparse.ArgumentParser("""
        This script loads images from annotations directory and trains ML to classify colony growth.
        """)
    parser.add_argument("annotations", type=str,
                        help="Directory containing annotated images")
    parser.add_argument("-v", "--visualise", action="store_true",
                        help="Generate visualisations for model diagnostics")
    parser.add_argument("-s", "--save", type=str,
                        help="If specified, tensorflow model will be saved to this folder")
    parser.add_argument("-l", "--log", action="store_true",
                        help="Store performance log in output folder")
    parser.add_argument("-t", "--test_dataset", type=str,
                        help="Testing dataset for final model evaluation. Ideally unseen data."
                             "If not provided then input directory is used (whole dataset).")
    parser.add_argument("-m", "--model_type", type=str, choices=["binary", "softmax"],
                        default="binary", help="Model type to train [default] = binary")
    parser.add_argument("-c", "--complexity", type=int, default=0, choices=[0, 1],
                        help="Model complexity level [default] = 0. Supports 0 or 1. "
                             "0 = first-step model from Gerada et al Spectrum 2024;"
                             "1 = second-step model from Gerada et al Spectrum 2024.")
    parser.add_argument("-b", "--batch_size", type=int, default=64,
                        help="Batch size for training [default] = 64")
    parser.add_argument("-d", "--dimensions", type=int, nargs=2, default=[160, 160],
                        help="X and Y dimensions of images for model training [default = 160 160]")
    return parser


def main():
    parser = train_modular_parser()
    args = parser.parse_args()

    image_width = args.dimensions[0]
    image_height = args.dimensions[1]

    annotated_images = pathlib.Path(args.annotations)
    print(f"Number of .jpg files{len(list(annotated_images.glob('*/*.jpg')))}")

    model_designs = (model_design_spectrum_2024_binary_first_step,
                     model_design_spectrum_2024_binary_second_step)

    if args.model_type == "binary":
        model, classes, history, _ = train_binary(annotations_path=annotated_images,
                                                  model_design=model_designs[args.complexity](image_width,
                                                                                              image_height),
                                                  image_width=image_width,
                                                  image_height=image_height,
                                                  batch_size=args.batch_size)
    elif args.model_type == "softmax":
        model, classes, history, _ = train_softmax(annotations_path=annotated_images,
                                                   model_design=model_designs[args.complexity](image_width,
                                                                                               image_height),
                                                   image_width=image_width,
                                                   image_height=image_height,
                                                   batch_size=args.batch_size)
    else:
        raise ValueError("Model type not supported")

    if args.save:
        if not os.path.exists(args.save):
            os.mkdir(args.save)
        model.save(args.save)
        print(f"Model saved to {args.save}")

    if args.log:
        if not args.save:
            warnings.warn(
                "Unable to save log file because model save path not provided, please use -s to provide path.")
        else:

            training_log_file = os.path.join(args.save, "training_log.csv")
            save_training_log(history, training_log_file)

    if args.test_dataset:
        if args.save is not None:
            annotation_log_file = os.path.join(args.save, "test_dataset_log.csv")
        else:
            annotation_log_file = None
        annotation_log = predict_colony_images_from_directory(args.test_dataset, model, classes, image_width,
                                                              image_height, args.model_type,
                                                              save_path=annotation_log_file)
        if args.visualise:
            for i in annotation_log:
                if i["predicted_class"] != i["true_class"]:
                    print(
                        f"This image was misclassified as {i['predicted_class']} "
                        f"with prediction of {i['prediction']} "
                        f"(should have been {i['true_class']})")
                    cv2.imshow(str(i['path']), i['image'])
                    cv2.waitKey()
                else:
                    print(f"Correct classification with prediction {i['prediction']} (class {i['true_class']})")

    if args.visualise:
        visualise_training(history)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# Filename: 	rename_images.py
# Author: 	Alessandro Gerada
# Date: 	2023-03-23
# Copyright: 	Alessandro Gerada 2023
# Email: 	alessandro.gerada@liverpool.ac.uk

"""Script to batch the renaming of a set of agar dilution images. Use rename_images.py -h for help."""

from aigarmic._img_utils import rename_jpg
import argparse
from argparse import RawTextHelpFormatter
import os


def rename_images_parser():
    parser = argparse.ArgumentParser(
        description='Script to batch rename images with halving value of drug concentration.',
        formatter_class=RawTextHelpFormatter)
    parser.add_argument('directory', type=str,
                        help="""
                            Directory containing images.
                            Filenames should reflect the concentration order when sorted, e.g.:
                                - image1 (no drug)
                                - image2 (low concentration)
                                - image3 (medium concentration)
                                - image4 (high concentration)
                            etc..
                            """)
    parser.add_argument('starting_concentration', type=float,
                        help='Starting (max) concentration that will be iteratively halved')
    return parser


def main():
    parser = rename_images_parser()
    args = parser.parse_args()

    files = os.listdir(args.directory)
    files = [i for i in files if i.lower().count(".jpg") > 0]
    
    if not files: 
        raise ValueError("No files found.")

    files.sort()
    control_plate = files.pop(0)
    rename_jpg(os.path.join(args.directory, control_plate),
               os.path.join(args.directory, '0.jpg'))

    concentration = args.starting_concentration

    while files: 
        _file = files.pop()
        rename_jpg(os.path.join(args.directory, _file),
                   os.path.join(args.directory, f"{concentration}.jpg"))
        concentration /= 2

if __name__ == '__main__':
    main()



# Filename: 	plate.py
# Author: 	Alessandro Gerada
# Date: 	2023-01-27
# Copyright: 	Alessandro Gerada 2023
# Email: 	alessandro.gerada@liverpool.ac.uk

"""Class implementation for plates"""
from pathlib import Path
from aigarmic.process_plate_image import split_by_grid
from aigarmic.model import Model
from aigarmic._img_utils import get_image_paths
from aigarmic.file_handlers import get_concentration_from_path
from typing import Optional, Union
import cv2  # pylint: disable=import-error
from random import randrange
import numpy as np
import os
from string import ascii_uppercase
from warnings import warn


class Plate:
    def __init__(self, drug: str,
                 concentration: float,
                 image_path: Optional[str] = None,
                 n_row: Optional[int] = None,
                 n_col: Optional[int] = None,
                 growth_code_matrix: Optional[list[list[int]]] = None,
                 visualise_contours: bool = False,
                 model: Optional[Model] = None,
                 key: Optional[list[str]] = None) -> None:
        """
        Store and process an agar plate image

        :param drug: Antibiotic name
        :param concentration: Antibiotic concentration
        :param image_path: Path to the image
        :param n_row: Number of rows in the plate
        :param n_col: Number of columns in the plate
        :param visualise_contours: Visualise the contours of the plate (useful for validation of grid splitting)
        :param model: Model to use for predictions
        :param key: Key to interpret model output (try to infer from model if not provided)
        """
        self.drug = drug
        self.concentration = concentration
        self.image_path = image_path

        if image_path is not None:
            if n_row is None or n_col is None:
                raise ValueError("Plate dimensions must be provided if image path is provided")
        self.n_row = n_row
        self.n_col = n_col

        self.accuracy_matrix = None
        self.model = model
        self.key = None

        if key is not None:
            if self.model is not None and self.model.get_key() != key:
                warn(f"Key provided to Plate does not match linked model key: {key} vs {self.model.get_key()}")
                warn(f"Plate will be using key parameter: {key}")
            self.key = key
        else:
            if self.model is not None:
                try:
                    self.key = self.model.get_key()
                except LookupError:
                    warn(f"No key found for linked model: {self.model}")

        if growth_code_matrix is not None:
            self.add_growth_code_matrix(growth_code_matrix)
        else:
            self.growth_code_matrix = None

        self.growth_matrix = None
        self.score_matrix = None
        self.predictions_matrix = None
        self.image_matrix = None
        self.image = None
        self.model_image_x = None
        self.model_image_y = None
        
        if self.image_path: 
            self.image = cv2.imread(self.image_path)
            self.image_matrix = split_by_grid(self.image, self.n_row, visualise_contours=visualise_contours,
                                              plate_name=self.drug + '_' + str(self.concentration))

    def add_growth_code_matrix(self, growth_code_matrix: list[list[int]]) -> None:
        if not self.valid_growth_code_matrix(growth_code_matrix):
            raise ValueError("Invalid growth code matrix, please provide a 2D growth coe matrix (list), values"
                             "must be integers, and cannot be negative")
        self.growth_code_matrix = growth_code_matrix
        dim_x, dim_y = self.matrix_dimensions(growth_code_matrix)
        if self.n_row is not None or self.n_col is not None:
            if dim_x != self.n_row or dim_y != self.n_col:
                raise ValueError(f"Dimensions of growth code matrix do not match plate dimensions: "
                                 f"{dim_x}x{dim_y} vs {self.n_row}x{self.n_col}")
        else:
            self.n_row = dim_x
            self.n_col = dim_y

    def valid_growth_code_matrix(self, growth_code_matrix: list[list[int]]) -> bool:
        try:
            self.matrix_dimensions(growth_code_matrix)
        except ValueError:
            return False

        for i in growth_code_matrix:
            for j in i:
                if not isinstance(j, int):
                    return False
                if self.key is not None:
                    if j > len(self.key) - 1:
                        return False
                if j < 0:
                    return False
        return True

    @staticmethod
    def matrix_dimensions(matrix) -> tuple[int, ...]:
        """
        Get dimensions of a matrix

        :param matrix: matrix to get dimensions of
        :raises ValueError: if matrix is not 2D or is not a valid matrix
        :return: tuple of dimensions
        """
        try:
            dimensions = np.shape(matrix)
            if len(dimensions) != 2:
                raise ValueError(f"Matrix {matrix} is not a 2D matrix")
            else:
                return dimensions
        except ValueError as e:
            raise ValueError(f"Matrix {matrix} is not a valid matrix") from e

    def split_images(self, visualise_contours: bool = False) -> None:
        """
        Splits images into individual colony images using grid

        :param visualise_contours: Visualise the contours of the plate (useful for validation of grid splitting)
        """
        self.image_matrix = split_by_grid(self.image, self.n_row,
                                          visualise_contours=visualise_contours,
                                          plate_name=self.drug + '_' + str(self.concentration))

    def import_image(self, image: np.ndarray) -> None:
        """
        Import and save image of agar plate

        :param image: loaded using cv2.imread
        """
        self.image = image

    def get_colony_image(self, index: Optional[tuple[int, int]] = None) -> tuple[np.ndarray, str]:
        """
        Pulls colony image and associated code-stamp
        Code-stamps are strings containing, in sequence:
        - Antibiotic name
        - Antibiotic concentration
        - Row (i) index
        - Column (j) index

        If no index is provided (default) a random image is given

        @param index: tuple of row and column index
        @return: tuple of image and code-stamp (e.g., "drug_0.125_i_1_j_2")
        """
        if index is None:
            i = randrange(self.n_row)
            j = randrange(self.n_col)
            image = self.image_matrix[i][j]
        else:
            try:
                i, j = index
                image = self.image_matrix[i][j]
            except KeyError as e:
                raise KeyError(f"Invalid index provided to get_colony_image: {index}") from e
        code = self.drug + "_" + str(self.concentration) + "_i_" + str(i) + "_j_" + str(j)
        return image, code

    def link_model(self, model: Model) -> None:
        """
        Link model to plate for predictions

        :param model: Model to link
        """
        self.model = model
        self.model_image_x = model.trained_x
        self.model_image_y = model.trained_y

    def get_key(self) -> Optional[list[str]]:
        """
        Get key from linked Model

        :raises: LookupError: No linked model to get key from
        :return: Key (or None if one is not found)
        """
        if self.key is not None:
            return self.key
        elif self.model is not None:
            return self.model.get_key()
        else:
            return None

    def set_key(self, key: list[str]) -> None:
        """
        Set plate key. Checks whether differs from linked model key (if any),
        and warns if different.

        :param key: List of growth categories (zero-indexed)
        """
        if self.model is not None:
            if self.model.get_key() != key:
                warn(f"Key provided to Plate does not match linked model key: {key} vs {self.model.get_key()}")
                warn(f"Plate will be overriding key parameter: {key}")
        self.key = key

    def annotate_images(self, model: Optional[Model] = None) -> list[list[str]]:
        """
        Annotate plate images

        :param model: linked model to use for predictions
        :return: Two-dimensional list of growth annotations
        """
        if not self.image_matrix:
            raise LookupError(
                """
                Unable to find an image_matrix associated with this plate. 
                Please provide an image path on construction or use import_image and split_images()
                """)
        model = self.model if not model else model 
        if not model: 
            raise LookupError(
                """
                Unable to find an image model for predictions associated with this plate. 
                Please provide one or use link_model(). 
                """
            )
        self.predictions_matrix = []
        self.score_matrix = []
        self.growth_matrix = []
        self.growth_code_matrix = []
        self.accuracy_matrix = []
        temp_score_row = []
        temp_predictions_row = []
        temp_growth_rows = []
        temp_growth_code_rows = []
        temp_accuracy_row = []
        for row in self.image_matrix: 
            for image in row: 
                prediction_data = model.predict(image)
                if 'growth_code' not in prediction_data:
                    raise ValueError("Model predictions dict must contain 'growth_code'")
                temp_predictions_row.append(prediction_data.get('prediction', None))
                temp_score_row.append(prediction_data.get('score', None))
                temp_growth_code_rows.append(prediction_data['growth_code'])
                _growth = None
                if 'growth' not in prediction_data:
                    try:
                        key = model.get_key()
                        _growth = key[prediction_data['growth_code']]
                    except LookupError:
                        pass
                temp_growth_rows.append(_growth)
                temp_accuracy_row.append(prediction_data.get('accuracy', 1.))
            self.predictions_matrix.append(temp_predictions_row)
            self.score_matrix.append(temp_score_row)
            self.growth_matrix.append(temp_growth_rows)
            self.growth_code_matrix.append(temp_growth_code_rows)
            self.accuracy_matrix.append(temp_accuracy_row)
            temp_score_row = []
            temp_predictions_row = []
            temp_growth_rows = []
            temp_accuracy_row = []
            temp_growth_code_rows = []
        return self.growth_matrix

    def print_matrix(self) -> None:
        """
        Print growth matrix in human-readable format
        """
        if not self.growth_matrix: 
            print(f"Plate {self.drug} - {self.concentration} not annotated")
        else: 
            for row in self.growth_matrix: 
                for result in row: 
                    print(str(result), sep="", end="")
                    print(" ", end="", sep="")
                print()

    def get_inaccurate_images(self, threshold: float = .9) -> set[tuple[int, int]]:
        """
        Get indexes of images with prediction accuracy below threshold
        :param threshold: Prediction threshold
        :return: Set containing indices of inaccurate images
        """
        output = set()
        for i, row in enumerate(self.accuracy_matrix):
            for j, item in enumerate(row):
                if item < threshold: 
                    output.add((i, j))
        return output
    
    def review_poor_images(self, threshold: float = .9,
                           save_dir: str = None) -> list[tuple[int, int]]:
        """
        Review and re-classify images with prediction accuracy below threshold. Classes should be zero indexed (e.g.,
        0, 1, 2). Currently, only supports up to 9.
        If save_dir provided then colony images will also be saved to a subdirectory (named after the new
        classification), to allow for future use in training.

        Enter new classification for each image using numbers (e.g., 0/1/2) on keyboard, press enter to skip,
        press esc to stop reviewing the plate.

        :param threshold: Prediction threshold to identify inaccurate images
        :param save_dir: Directory to save re-classified images
        :return: List of indices of re-classified images
        """
        codes = {}
        for ascii_code, class_code in zip(range(48, 58), range(0, 10)):
            codes[ascii_code] = class_code
        skip_codes = {13: "enter"}
        stop_codes = {27: "esc"}
        codes.update(skip_codes)
        codes.update(stop_codes)

        inaccurate_images_indexes = self.get_inaccurate_images(threshold) 
        changed_log = []
        for image_index in inaccurate_images_indexes:
            image, stamp = self.get_colony_image(image_index)
            i, j = image_index
            growth = self.growth_matrix[i][j]
            accuracy = self.accuracy_matrix[i][j]
            print()
            print(f"This image ({self.drug + str(self.concentration)} position {i} {j}) was labelled as {growth} "
                  f"with an accuracy of {accuracy * 100:.2f}")
            cv2.imshow(self.drug + str(self.concentration) + f" position {i} {j}", image)
            print("Press enter to continue, or enter new classification: ")
            while True: 
                input_key = cv2.waitKey()
                if input_key not in codes:
                    print("Input not recognised, please try again..")
                    continue
                if input_key in stop_codes or input_key in skip_codes:
                    break
                try:
                    _ = self.get_key()[codes[input_key]]
                except IndexError:
                    print(f"Invalid input {codes[input_key]}: model key is {self.get_key()} [zero-indexed]")
                    continue
                break

            input_code = codes[input_key]

            if input_code in stop_codes.values():
                print(f"Stopping review for this plate: {self}.")
                break

            if input_code in skip_codes.values():
                print("Classification not changed.")
                continue

            if self.get_key()[input_code] == growth:
                print("Classification unchanged.")
                continue

            # reassign growth
            print(f"Reassigning image to {self.get_key()[input_code]}")
            self.growth_matrix[i][j] = self.get_key()[input_code]
            self.growth_code_matrix[i][j] = input_code
            changed_log.append(image_index)
            if save_dir:
                if not os.path.exists(save_dir):
                    print(f"Creating directory: {save_dir}")
                    os.mkdir(save_dir)
                class_dir = os.path.join(save_dir, str(input_code))
                if not os.path.exists(class_dir):
                    print(f"Creating class subdirectory: {class_dir}")
                    os.mkdir(class_dir)
                save_path = os.path.join(class_dir, stamp + ".jpg")
                print(f"Saving image to: {save_path}")
                cv2.imwrite(save_path, image)
        return changed_log

    def convert_growth_codes(self, key: list[str]) -> list[list[str]]:
        """
        Convert growth codes to human-readable format using key
        E.g., [0, 1, 2] -> ["No growth", "Poor growth", "Good growth"]
        Sets self.growth_matrix

        :param key: List of growth codes (zero-indexed)
        :return: Growth matrix
        """
        self.growth_matrix = []
        for row in self.growth_code_matrix:
            _temp_row = []
            for code in row:
                _temp_row.append(key[code])
            self.growth_matrix.append(_temp_row)
        return self.growth_matrix

    def __repr__(self) -> str:
        return f"Plate of {self.drug} at {self.concentration}mg/L"

    def __lt__(self, other) -> bool:
        return self.concentration < other.concentration

    def __eq__(self, other) -> bool:
        return self.concentration == other.concentration and self.drug == other.drug

    def __gt__(self, other) -> bool:
        return self.concentration > other.concentration

    def __le__(self, other) -> bool:
        return self.concentration <= other.concentration

    def __ge__(self, other) -> bool:
        return self.concentration >= other.concentration

    def __ne__(self, other) -> bool:
        return self.concentration != other.concentration

    def __hash__(self) -> int:
        return hash((self.drug, self.concentration))


class PlateSet: 
    def __init__(self, plates_list: list[Plate],
                 key: Optional[list[str]] = None) -> None:
        """
        Combines a list of Plate objects into a PlateSet to facilitate MIC calculation.
        Generally, plates would have a range of antimicrobial concentrations, including a control plate (concentration
        of 0.0mg/L).
        Plates must be annotated before initialisation (using Plate.annotate_images()) and have the same antimicrobial
        name and growth keys.

        :param plates_list: List of Plate objects
        """
        self.no_growth_key_items = None
        self.qc_matrix = None
        self.mic_matrix = None

        if key is not None:
            self.key = key
        else:
            _list_of_keys = []
            try:
                _list_of_keys = [i.get_key() for i in plates_list]
                self.key = _list_of_keys[0]
            except LookupError:
                warn(f"No key provided to PlateSet")
            if not all(i == _list_of_keys[0] for i in _list_of_keys):
                raise ValueError("Plates supplied to PlateSet have different growth keys")
            if not _list_of_keys:
                self.key = None

        drug_names = [i.drug for i in plates_list]
        if len(set(drug_names)) > 1:
            raise ValueError("Plates supplied to PlateSet have different antibiotic names")
        elif not len(set(drug_names)):
            raise ValueError("Plates supplied to PlateSet do not have antibiotic names")

        self.drug = plates_list[0].drug
        self.antibiotic_plates = [i for i in plates_list if i.concentration != 0.0]

        _temp_positive_control_plate = [i for i in plates_list if i.concentration == 0.0]
        if not _temp_positive_control_plate:
            warn(f"No control plate supplied to {self.drug} PlateSet")
        if len(_temp_positive_control_plate) > 1:
            warn(f"Multiple control plates supplied to {self.drug} PlateSet, control plates will be skipped.")
        else:
            [self.positive_control_plate] = _temp_positive_control_plate

        self.antibiotic_plates = sorted(self.antibiotic_plates)

        # check dimensions of plates' matrices
        if not self.valid_dimensions(): 
            raise ValueError("Plate matrices have different dimensions - unable to calculate MIC")

    def valid_dimensions(self) -> bool:
        """
        Check if all plates in PlateSet have the same x and y dimensions

        :return: True if all plates have the same dimensions, False otherwise
        """
        matrices_shapes = []
        for i in self.get_all_plates():
            try:
                matrices_shapes.append(i.matrix_dimensions(i.growth_code_matrix))
            except ValueError as e:
                raise ValueError(f"Plate {i} does not a matrix-shaped growth code matrix") from e

        return True if all(i == matrices_shapes[0] for i in matrices_shapes) else False

    def get_all_plates(self) -> list[Plate]:
        """
        Returns a sorted list of all plates in the PlateSet, including the control plate

        :return: List of Plate objects
        """
        return sorted(self.antibiotic_plates + [self.positive_control_plate])
    
    def convert_mic_matrix(self, mic_format: str = "string") -> np.array:
        """
        Converts format of MIC matrix

        :param mic_format: Format to convert to (only "string" is supported)
        :return: matrix (array) of MIC values
        """
        allowed_formats = ["string"]
        format_conversion = {"string": str}

        if mic_format not in allowed_formats:
            raise ValueError(f"MIC matrix formats must be one of: {allowed_formats}")
        output = self.mic_matrix.astype(format_conversion[mic_format])
        if mic_format == "string":
            max_mic_plate = max([i.concentration for i in self.antibiotic_plates])
            min_mic_plate = min([i.concentration for i in self.antibiotic_plates])
            for i, row in enumerate(output):
                for j, mic in enumerate(row):
                    if float(mic) > max_mic_plate: 
                        output[i][j] = ">" + str(max_mic_plate)
                    elif float(mic) == min_mic_plate: 
                        output[i][j] = "<" + str(min_mic_plate)
                    else: 
                        output[i][j] = mic
        return output

    def calculate_mic(self, no_growth_key_items: tuple[int, ...]) -> np.array:
        """
        Calculate MIC matrix using image predictions.
        Sets self.mic_matrix

        :param no_growth_key_items: tuple of key items that should be classified as "no growth" for MIC purposes
        :return: MIC matrix
        """
        self.no_growth_key_items = no_growth_key_items
        self.antibiotic_plates = sorted(self.antibiotic_plates, reverse=True)
        max_concentration = max([i.concentration for i in self.antibiotic_plates])*2
        mic_matrix = np.array(self.antibiotic_plates[0].growth_code_matrix)
        mic_matrix = np.full(mic_matrix.shape, max([i.concentration for i in self.antibiotic_plates])*2)
        rows = range(mic_matrix.shape[0])
        cols = range(mic_matrix.shape[1])

        def get_first_negative_concentration(starting_concentration, i, j):
            if self.antibiotic_plates[0].growth_code_matrix[i][j] not in self.no_growth_key_items:
                return starting_concentration
            c = self.antibiotic_plates[0].concentration
            for plate in self.antibiotic_plates[1:]:
                if plate.growth_code_matrix[i][j] not in self.no_growth_key_items:
                    return c
                else:
                    c = plate.concentration
            return c
            
        for row in rows:
            for col in cols:
                mic_matrix[row][col] = get_first_negative_concentration(max_concentration, row, col)
        self.mic_matrix = mic_matrix
        return mic_matrix

    def generate_qc(self) -> np.array:
        """
        Generate QC matrix for PlateSet, as follows:

        "F" = FAIL - no growth in positive control plate, result should be disregarded
        "W" = WARNING - more than one change in concentration gradient. There should only be one change at the MIC
        breakpoint (where the images change from growth to no/poor growth). Depending on the application of the results,
        manual confirmation should be considered for warnings.
        "P" = PASS - no QC issues found

        :return: Matrix of QC values (strings)
        """
        if self.mic_matrix is None:
            raise ValueError(f"MIC matrix not found for {repr(self)} - please calculate MIC using calculate_mic()")
        qc_matrix = np.full(self.mic_matrix.shape, fill_value="", dtype=str)

        if self.positive_control_plate is None:
            warn(f"*Warning* - {repr(self)} does not contain a positive control plate.")
        else:
            for i, row in enumerate(self.positive_control_plate.growth_code_matrix):
                for j, item in enumerate(row):
                    if item in self.no_growth_key_items:
                        qc_matrix[i][j] = "F"
                    else:
                        qc_matrix[i][j] = "P"

        def simplify_codes(code):
            return 0 if code in self.no_growth_key_items else 1
        
        antibiotic_plates = sorted(self.antibiotic_plates, reverse=True)
        if len(antibiotic_plates) > 1: 
            rows = range(qc_matrix.shape[0])
            cols = range(qc_matrix.shape[1])
            for i in rows: 
                for j in cols:
                    if qc_matrix[i][j] == "F":
                        continue
                    previous_growth_code = simplify_codes(antibiotic_plates[0].growth_code_matrix[i][j])
                    flips = 0  # we only allow one "flip" from no growth -> growth
                    for k in antibiotic_plates[1:]:
                        next_growth_code = simplify_codes(k.growth_code_matrix[i][j])
                        if next_growth_code != previous_growth_code:
                            flips += 1
                        previous_growth_code = next_growth_code
                    if flips > 1:
                        qc_matrix[i][j] = "W"
        else:
            warn(f"*Warning* - {repr(self)} has insufficient plates for full QC")
        self.qc_matrix = qc_matrix
        return qc_matrix

    def review_poor_images(self, threshold: float = .9,
                           save_dir: Optional[str] = None) -> list[list[tuple[int, int]]]:
        """
        Review and re-classify images with prediction accuracy below threshold. Currently, only supports three levels,
        0 (no growth), 1 (poor growth), 2 (good growth), as described in Gerada et al. (2024), Microbiology Spectrum.
        If save_dir provided then colony images will also be saved to a subdirectory (named after the new
        classification), to allow for future use in training.

        Enter new classification for each image using 0/1/2 on keyboard, or press enter (or esc) to skip.

        :param threshold: Prediction threshold to identify inaccurate images
        :param save_dir: Directory to save re-classified images
        :return: List of indices of re-classified images
        """
        changed = [i.review_poor_images(threshold, save_dir) for i in self.get_all_plates()]
        n_changed = 0
        for i in changed:
            n_changed += len(i)
        print(f"{n_changed} images re-classified.")
        return changed

    def get_csv_data(self) -> list[dict]:
        """
        Get MIC and QC data in a format suitable for CSV export:
        List of dicts containing:
        - Antibiotic: Antibiotic name
        - Position: Position of the colony (e.g., A1, B2, etc.)
        - MIC: MIC value
        - QC: QC value (P, W, F)

        :return: List of dicts with MIC and QC data
        """
        if self.mic_matrix is None:
            raise ValueError("Please calculate MIC using PlateSet.calculate_mic() before exporting data")
        if self.qc_matrix is None:
            raise ValueError("Please generate QC using PlateSet.generate_qc() before exporting data")

        mic_matrix_str = self.convert_mic_matrix(mic_format="string")
        row_letters = ascii_uppercase[0:len(mic_matrix_str)]
        col_nums = [i + 1 for i in range(len(mic_matrix_str[0]))]
        output = []
        for i in range(len(row_letters)):
            for j in range(len(col_nums)):
                position = row_letters[i]+str(col_nums[j])
                mic = mic_matrix_str[i][j]
                qc = self.qc_matrix[i][j]
                output.append({'Antibiotic': self.drug, 'Position': position, 'MIC': mic, 'QC': qc})
        return output
        
    def __repr__(self) -> str:
        return (f"PlateSet of {self.drug} with {len(self.antibiotic_plates)} "
                f"concentrations: {[i.concentration for i in self.antibiotic_plates]}")


def plate_set_from_dir(path: Union[str, Path],
                       drug: str,
                       model: Model,
                       n_row: int = 8,
                       n_col: int = 12,
                       **kwargs) -> PlateSet:
    """
    Create a PlateSet from a directory of images. Images are annotated using the provided model.

    :param path: directory containing plate images (.jpg) with filenames indicating antibiotic concentration
    :param drug: name of drug
    :param model: model file to use for predictions
    :param n_row: number of rows in the plates
    :param n_col: number of columns in the plates
    :param kwargs: additional keyword arguments to pass to Plate constructor
    :return: PlateSet with MIC and QC values
    """
    image_paths = get_image_paths(path)
    plates = [Plate(drug,
                    concentration=get_concentration_from_path(i),
                    image_path=i,
                    model=model,
                    n_row=n_row,
                    n_col=n_col,
                    **kwargs) for i in image_paths]
    for i in plates:
        i.annotate_images()
    output = PlateSet(plates)
    return output



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Filename: 	model_performance.py
# Author: 	Alessandro Gerada
# Date: 	2023-03-16
# Copyright: 	Alessandro Gerada 2023
# Email: 	alessandro.gerada@liverpool.ac.uk

"""Test the performance of a saved model on a set of colony images. Use model_performance.py -h for help."""

from aigarmic.file_handlers import get_paths_from_directory
from aigarmic.model import BinaryModel
import cv2  # pylint: disable=import-error
import argparse


def model_performance_parser():
    parser = argparse.ArgumentParser(description="Evaluate saved model performance")
    parser.add_argument("model_path", type=str, help="Path to the model")
    parser.add_argument("annotations_path", type=str, help="Path to the annotations to test")
    parser.add_argument("-x", "--trained_x", type=int, default=160,
                        help="Width of the images that the model was trained on [default=160]")
    parser.add_argument("-y", "--trained_y", type=int, default=160,
                        help="Height of the images that the model was trained on [default=160]")
    parser.add_argument("-t", "--threshold", type=float, default=0.95,
                        help="Accuracy threshold, to highlight low accuracy predictions [default=0.95]")
    return parser


def main():
    parser = model_performance_parser()
    args = parser.parse_args()

    images = get_paths_from_directory(args.annotations_path)
    model = BinaryModel(args.model_path,
                        trained_x=args.trained_x,
                        trained_y=args.trained_y,
                        key=['No growth', 'Growth'])

    results = []
    for i, paths in images.items():
        for path in paths:
            image = cv2.imread(path)
            p = model.predict(image)
            p['true_class'] = int(i)
            results.append(p)

    color_end = '\033[0m'
    color_red = '\033[91m'
    color_green = '\033[92m'

    errors = 0
    for i in results:
        if i['growth_code'] != i['true_class']:
            errors += 1
            color = color_red
        else:
            color = color_green
        print(f"Prediction={i['prediction']:.2f} \t "
              f"Growth code={color}{i['growth_code']}{color_end} \t "
              f"True class={color}{i['true_class']}{color_end} \t "
              f"Accuracy={color_red if i['accuracy'] < args.threshold else color_green}{i['accuracy']:.2f}{color_end}")

    print()
    print(f"Errors: {errors} from {len(results)} images ({errors/len(results)*100:.2f}%)")

    threshold = 0.9
    true_positives = 0
    false_positives = 0
    true_negatives = 0
    false_negatives = 0
    n = len(results)
    for i in results:
        if i['growth_code'] != i['true_class']:
            if i['accuracy'] < threshold:
                true_positives += 1
            else:
                false_negatives += 1
        else:
            if i['accuracy'] < threshold:
                false_positives += 1
            else:
                true_negatives += 1

    print(f"TP={true_positives} ({true_positives/n*100:.2f}%)")
    print(f"FP={false_positives} ({false_positives/n*100:.2f}%)")
    print(f"TN={true_negatives} ({true_negatives/n*100:.2f}%)")
    print(f"FN={false_negatives} ({false_negatives/n*100:.2f}%)")


if __name__ == "__main__":
    main()



# Filename:      train.py
# Author:        Alessandro Gerada
# Date:          21/03/2024
# Copyright:     Alessandro Gerada 2024
# Email:         alessandro.gerada@liverpool.ac.uk

"""
Functions and classes that allow for training neural network models for colony image classification.
"""

from aigarmic.file_handlers import create_dataset_from_directory
import tensorflow as tf
from tensorflow.keras.models import Sequential
from tensorflow.keras import layers
from tensorflow.keras.utils import image_dataset_from_directory
from tensorflow.keras.initializers import Constant
import keras.callbacks
from sklearn.utils import class_weight
from tensorflow.keras import initializers
from pathlib import Path
import numpy as np
import matplotlib.pyplot as plt


def train_binary(annotations_path,
                 model_design: Sequential,
                 val_split: float = 0.2,
                 image_width: int = 160,
                 image_height: int = 160,
                 batch_size: int = 64,
                 epochs: int = 300,
                 stop_training_threshold: float = 0.98,
                 learning_rate: float = .0001) -> tuple[Sequential, list[str], keras.callbacks.History, list]:
    """
    Train a binary classification model to differentiate between two classes of colony images.
    Provide a keras sequential model design to inform the neural network architecture. The final binary/sigmoid layer
    should not be included in the sequential model design, as this is added by the function.

    :param annotations_path: Path to directory containing annotated images, with subdirectories for each class (usually
        '0' and '1')
    :param model_design: Keras model design (Sequential) to inform neural network architecture, excluding final layer
    :param val_split: Validation split proportion
    :param image_width: Image width (pixels)
    :param image_height: Image height (pixels)
    :param batch_size: Training batch size
    :param epochs: Max number of training epochs
    :param stop_training_threshold: Accuracy threshold at which to accept model and stop training
    :param learning_rate: Learning rate for the optimizer

    :return: Tuple containing trained model, class names, training history, and evaluation results
    """
    label_mode = "binary"
    train_dataset, val_dataset = create_dataset_from_directory(
        directory=annotations_path,
        label_mode=label_mode,
        image_width=image_width,
        image_height=image_height,
        val_split=val_split,
        batch_size=batch_size
    )

    class_names = train_dataset.class_names
    num_classes = len(class_names)
    if num_classes != 2:
        raise ValueError("Binary classification requires 2 classes")

    full_dataset = image_dataset_from_directory(Path(annotations_path),
                                                image_size=(image_height, image_width),
                                                batch_size=batch_size,
                                                label_mode="binary")

    full_dataset_unbatched = tuple(full_dataset.unbatch())
    labels = [0, 0]
    for (_, label) in full_dataset_unbatched:
        labels[int(label.numpy())] += 1
    neg = labels[0]
    pos = labels[1]
    initial_bias = np.log([pos / neg])
    initial_bias = Constant(initial_bias)

    weight_0 = (1 / neg) * ((neg + pos) / 2)
    weight_1 = (1 / pos) * ((neg + pos) / 2)
    class_weights = {0: weight_0, 1: weight_1}

    model_design.add(layers.Dense(1, activation='sigmoid',
                                  bias_initializer=initial_bias))
    with tf.device('/device:GPU:0'):
        model_design.compile(optimizer=tf.keras.optimizers.legacy.Adam(learning_rate=learning_rate),
                             loss='binary_crossentropy',
                             metrics=['accuracy'])

    val_callback = ValidationThresholdCallback(threshold=stop_training_threshold)
    history = model_design.fit(
        train_dataset,
        validation_data=val_dataset,
        epochs=epochs,
        class_weight=class_weights,
        callbacks=[val_callback]
    )
    results = model_design.evaluate(train_dataset, batch_size=batch_size, verbose=0)

    return model_design, class_names, history, results


def train_softmax(annotations_path,
                  model_design: Sequential,
                  val_split: float = 0.2,
                  image_height: int = 160,
                  image_width: int = 160,
                  batch_size: int = 64,
                  epochs: int = 300,
                  stop_training_threshold: float = 0.98) -> tuple[Sequential, list[str], keras.callbacks.History, list]:
    """
    Train a softmax classification model to differentiate between multiple classes (2 or more) of colony images.
    The final softmax layer should not be included in the sequential model design, as this is added by the function.

    :param annotations_path: Path to directory containing annotated images, with subdirectories for each class
        (e.g., '0', '1', '2', ...)
    :param model_design: Keras model design (Sequential) to inform neural network architecture, excluding final layer
    :param val_split: Validation split proportion
    :param image_width: Image width (pixels)
    :param image_height: Image height (pixels)
    :param batch_size: Training batch size
    :param epochs: Max number of training epochs
    :param stop_training_threshold: Accuracy threshold at which to accept model and stop training

    :return: Tuple containing trained model, class names, training history, and evaluation results
    """
    label_mode = "int"
    train_dataset, val_dataset = create_dataset_from_directory(
        directory=annotations_path,
        label_mode=label_mode,
        image_width=image_width,
        image_height=image_height,
        val_split=val_split,
        batch_size=batch_size
    )
    class_names = train_dataset.class_names
    num_classes = len(class_names)
    obs_dict = dict(zip(list(range(num_classes)), [0] * num_classes))
    all_class_obs = []
    for _, j in train_dataset:
        for k in j:
            k_arr = k.numpy()
            obs_dict[k_arr] += 1
            all_class_obs.append(k_arr)

    class_weights = class_weight.compute_class_weight('balanced',
                                                      classes=np.unique(all_class_obs),
                                                      y=all_class_obs)
    class_weights = dict(zip(list(range(num_classes)), class_weights))
    initial_bias = initializers.Zeros()

    model_design.add(layers.Dense(num_classes, activation='softmax', bias_initializer=initial_bias))

    with tf.device('/device:GPU:0'):
        model_design.compile(optimizer='adam',
                             loss='sparse_categorical_crossentropy',
                             metrics=['accuracy'])

        val_callback = ValidationThresholdCallback(threshold=stop_training_threshold)
        history = model_design.fit(
            train_dataset,
            validation_data=val_dataset,
            epochs=epochs,
            class_weight=class_weights,
            callbacks=[val_callback]
        )

    results = model_design.evaluate(train_dataset, batch_size=batch_size, verbose=0)
    return model_design, class_names, history, results


def visualise_training(history: keras.callbacks.History) -> None:
    """
    Visualise training and validation accuracy and loss over epochs.

    :param history: Training history object (usually returned by model.fit(), train_binary(), or train_softmax())
    """
    epochs = len(history.history['accuracy'])
    acc = history.history['accuracy']
    val_acc = history.history['val_accuracy']

    loss = history.history['loss']
    val_loss = history.history['val_loss']

    epochs_range = range(epochs)

    plt.figure(figsize=(8, 8))
    plt.subplot(1, 2, 1)
    plt.plot(epochs_range, acc, label='Training Accuracy')
    plt.plot(epochs_range, val_acc, label='Validation Accuracy')
    plt.legend(loc='lower right')
    plt.title('Training and Validation Accuracy')

    plt.subplot(1, 2, 2)
    plt.plot(epochs_range, loss, label='Training Loss')
    plt.plot(epochs_range, val_loss, label='Validation Loss')
    plt.legend(loc='upper right')
    plt.title('Training and Validation Loss')
    plt.show()


class ValidationThresholdCallback(tf.keras.callbacks.Callback):
    def __init__(self, threshold):
        """
        Stop training if validation accuracy is above threshold

        :param threshold: Threshold to stop training
        """
        super().__init__()
        self.threshold = threshold

    def on_epoch_end(self, epoch, logs=None) -> None:
        """
        Determines whether to stop training based on validation accuracy
        """
        val_acc = logs["val_accuracy"]
        if val_acc >= self.threshold:
            self.model.stop_training = True



# Filename: 	process_plate_image.py
# Author: 	Alessandro Gerada
# Date: 	2023-01-27
# Copyright: 	Alessandro Gerada 2023
# Email: 	alessandro.gerada@liverpool.ac.uk

"""Pre-processing of agar plate images to extract individual colonies"""

import cv2  # pylint: disable=import-error
from imutils import contours
from typing import Optional
from numpy import ndarray, empty


def find_threshold_value(image: ndarray, start: int = 20,
                         end: int = 100,
                         by: int = 1,
                         look_for: int = 96,
                         area_lower_bound: int = 1000) -> Optional[tuple[list, int]]:
    """
    Find threshold value that correctly splits an agar plate image into colony sub-images. Assumes that a black grid
    overlays the image.
    :param image: Image file loaded using cv2.imread
    :param start: starting threshold value
    :param end: ending threshold value
    :param by: threshold increment value
    :param look_for: target sub-images
    :param area_lower_bound: minimum area for a contour to be considered
    :return: tuple of contours and threshold value
    """
    for i in range(start, end, by):
        ret, thresh = cv2.threshold(image, i, 255, cv2.THRESH_BINARY_INV)

        # Find contours and filter using area
        _contours = cv2.findContours(thresh, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
        _contours = _contours[0] if len(_contours) == 2 else _contours[1]
        grid_contours = []
        for c in _contours:
            area = cv2.contourArea(c)
            if area > area_lower_bound:
                grid_contours.append(c)

        # sort contours and remove biggest (outer) grid square
        grid_contours = sorted(grid_contours, key=cv2.contourArea)
        grid_contours = grid_contours[:-1]

        # If we find the target boxes, return contours and threshold
        if len(grid_contours) == look_for:
            return grid_contours, i
    return None


def split_by_grid(image: ndarray, n_rows: int = 8,
                  visualise_contours: bool = False,
                  plate_name: Optional[str] = None) -> list[list[ndarray]]:
    """
    Split an agar plate image into individual colony sub-images using a grid overlay.

    :param image: image file loaded using cv2.imread
    :param n_rows: number of rows in the grid (columns will be inferred automatically)
    :param visualise_contours: if True, display the contours found (useful for validation)
    :param plate_name: name of plate to display in visualisation (useful for validation)
    :return: matrix of sub-images
    """
    if visualise_contours and not plate_name:
        raise ValueError("Pass plate name to split_by_grid if using visualise_contours")
    blur = cv2.GaussianBlur(image, (25, 25), 0)
    gray = cv2.cvtColor(blur, cv2.COLOR_BGR2GRAY)
    grid_contours, threshold_value = find_threshold_value(gray)

    if visualise_contours:
        _image = image
        cv2.drawContours(_image, grid_contours, -1, (0, 255, 0), 10)
        cv2.imshow(plate_name, _image)
        cv2.waitKey()

    if not grid_contours:
        raise ValueError("Unable to find contours threshold that returns correct number of colony images")

    # Sort contours, starting left to right
    (grid_contours, _) = contours.sort_contours(grid_contours, method="left-to-right")
    sorted_grid = []
    col = []  # temporary list to hold columns while sorting

    for (i, c) in enumerate(grid_contours, 1):
        col.append(c)
        if i % n_rows == 0:
            # found column - sort top to bottom and add to output
            (c_tmp, _) = contours.sort_contours(col, method="top-to-bottom")
            sorted_grid.append(c_tmp)
            col = []

    out_matrix = [[empty(shape=1) for _ in range(len(sorted_grid))] for _ in range(n_rows)]

    # Iterate through each box
    for j, col in enumerate(sorted_grid):
        for i, c in enumerate(col):
            x, y, w, h = cv2.boundingRect(c)
            cropped_image = image[y:y + h, x:x + w]
            out_matrix[i][j] = cropped_image

    return out_matrix



# Filename: 	_nn_design.py
# Author: 	Alessandro Gerada
# Date: 	2023-08-11
# Copyright: 	Alessandro Gerada 2023
# Email: 	alessandro.gerada@liverpool.ac.uk

"""Template NN designs as reported in Gerada et al. 2024 Microbiology Spectrum paper"""

from tensorflow.keras import layers
from tensorflow.keras.models import Sequential


def model_design_spectrum_2024_binary_first_step(image_width: int,
                                                 image_height: int,
                                                 augmentation: bool = True):
    design = [
        #  spectrum 2024 first-step model
        layers.Rescaling(1./255, input_shape=(image_width, image_height, 3)),
        layers.Conv2D(32, (3,3), activation='relu'),
        layers.MaxPooling2D((2,2)),

        layers.Conv2D(32, (3,3), kernel_initializer='he_uniform', activation='relu'),
        layers.MaxPooling2D((2,2)),

        layers.Conv2D(64, (3,3), kernel_initializer='he_uniform', activation='relu'),
        layers.MaxPooling2D((2,2)),

        layers.Flatten(),
        layers.Dense(64, activation='relu'),
        layers.Dropout(0.1)
    ]

    data_augmentation = [
        layers.RandomFlip("horizontal",
                          input_shape=(image_width,
                                       image_height,
                                       3)),
        layers.RandomRotation(0.1),
        layers.RandomContrast(0.1),
        layers.RandomBrightness(0.1)
    ]
    if augmentation:
        design = data_augmentation + design
    print(design)
    return Sequential(design)


def model_design_spectrum_2024_binary_second_step(image_width: int,
                                                  image_height: int,
                                                  augmentation: bool = True):
    design = [
        #  spectrum 2024 second-step model
        layers.Rescaling(1. / 255, input_shape=(image_width, image_height, 3)),
        layers.Conv2D(128, (3, 3), activation='relu', padding='same'),
        layers.MaxPooling2D((2, 2)),
        layers.Flatten(),
        layers.Dense(128, activation='relu'),
        layers.Dropout(0.1),
        layers.Dense(128, activation='relu'),
        layers.Dropout(0.1)
    ]

    data_augmentation = [
        layers.RandomFlip("horizontal",
                          input_shape=(image_width,
                                       image_height,
                                       3)),
        layers.RandomRotation(0.1),
        layers.RandomContrast(0.1),
        layers.RandomBrightness(0.1)
    ]
    if augmentation:
        design = data_augmentation + design

    return Sequential(design)



# Filename: 	file_handlers.py
# Author: 	Alessandro Gerada
# Date: 	2023-08-11
# Copyright: 	Alessandro Gerada 2023
# Email: 	alessandro.gerada@liverpool.ac.uk

"""Functions to facilitate working with image data files"""

import pathlib
from pathlib import Path

from aigarmic._img_utils import convert_cv2_to_keras
import csv
import os
from typing import Optional, Union
import cv2  # pylint: disable=import-error
import keras.callbacks
import numpy as np
import tensorflow as tf


def create_dataset_from_directory(directory: str,
                                  label_mode: str,
                                  image_width: int,
                                  image_height: int,
                                  seed: int = 12345,
                                  val_split: float = 0.2,
                                  batch_size: int = 32) -> tuple[tf.data.Dataset, tf.data.Dataset]:
    """
    Create a training and validation dataset from a directory containing subdirectories for each class of image data.

    :param directory: path containing images, each in subdirectory corresponding to class
    :param label_mode: Labelling depending on model type ("binary" for binary or "int" for softmax)
    :param image_width: Image width in pixels
    :param image_height: Image height in pixels
    :param seed: Random seed for dataset splitting
    :param val_split: Proportion of data to use for validation
    :param batch_size: Batch size for datasets
    :return: Tuple containing training and validation datasets
    """
    train_dataset = tf.keras.utils.image_dataset_from_directory(
                directory,
                validation_split=val_split, 
                subset='training',
                seed=seed,
                image_size=(image_width, image_height), 
                batch_size=batch_size, 
                label_mode=label_mode
            )
    val_dataset = tf.keras.utils.image_dataset_from_directory(
        directory,
        validation_split=val_split, 
        subset='validation', 
        seed=seed,
        image_size=(image_width, image_height), 
        batch_size=batch_size, 
        label_mode=label_mode
    )
    return train_dataset, val_dataset


def predict_colony_images_from_directory(directory: Optional[Union[str, pathlib.Path]],
                                         model: tf.keras.models.Model,
                                         class_names: list[str],
                                         image_width: int,
                                         image_height: int,
                                         model_type: str,
                                         save_path: Optional[Union[str, pathlib.Path]] = None,
                                         binary_threshold: float = 0.5) -> list[dict]:
    """
    Predict the class of images in a directory using a trained model, and compare prediction to
    true class (based on subdirectory in which image is located, which should correspond to class name).

    :param directory: Directory containing images to predict
    :param model: Model to use for prediction
    :param class_names: List of class names
    :param image_width: Image width in pixels
    :param image_height: Image height in pixels
    :param model_type: "binary" or "softmax"
    :param save_path: Path to save prediction log
    :param binary_threshold: For binary models, threshold for classifying as positive
    :return: List of dictionaries containing image, path, prediction, predicted class, and true class (for each image)
    """
    output = []
    file_paths = {i: os.listdir(os.path.join(directory, i)) for i in class_names}
    # add subdirectories
    file_paths = {i: [os.path.join(directory, i, j) for j in file_paths[i] if j.count(".jpg") > 0] for i in file_paths}

    for i in file_paths:
        for j in file_paths[i]:
            image = cv2.imread(j)
            true_class = i
            directory = j
            prediction = model.predict(convert_cv2_to_keras(image, image_width, image_height))
            if model_type == "binary":
                [prediction] = prediction.reshape(-1)
                predicted_class = class_names[0] if prediction <= binary_threshold else class_names[1]
            elif model_type == "softmax":
                prediction = tf.nn.softmax(prediction)
                prediction_value = np.max(prediction)
                predicted_class = np.argmax(prediction)
                prediction = prediction_value
            else:
                raise ValueError("Model type not supported")
            output.append({"image": image,
                           "path": directory,
                           "prediction": prediction,
                           "predicted_class": predicted_class,
                           "true_class": true_class})
    if save_path:
        with open(save_path, "w", encoding="utf-8-sig") as file:
            writer = csv.DictWriter(file,
                                    fieldnames=['path', 'prediction', 'predicted_class', 'true_class'],
                                    extrasaction='ignore')
            writer.writeheader()
            writer.writerows(output)
    return output


def save_training_log(model_history: keras.callbacks.History,
                      save_path: Union[str, pathlib.Path]) -> None:
    """
    Save training log to CSV file

    :param model_history: Training history object
    :param save_path: Directory to save training log
    """
    with open(save_path, "w", encoding="utf-8-sig") as file:
        writer = csv.writer(file)
        h = zip(
            model_history.history['accuracy'],
            model_history.history['val_accuracy'],
            model_history.history['loss'],
            model_history.history['val_loss'],
            range(len(model_history.history['accuracy'])))
        writer.writerow(['accuracy', 'val_accuracy', 'loss', 'val_loss', 'epoch'])
        writer.writerows(h)


def get_concentration_from_path(path: Union[str, Path]) -> float:
    """
    get concentration from plate image path, e.g.
    antibiotic1/0.125.jpg -> 0.125

    :param path: Path to plate image
    :return: Concentration
    """
    split_text = os.path.split(path)
    split_text = split_text[-1]
    concentration_str = os.path.splitext(split_text)[0]
    return float(concentration_str)


def get_paths_from_directory(path: Union[str, Path]) -> dict[str, list[str]]:
    """
    Returns a dict of abx_names: [image1_path, image2_path, etc.]
    If there are no antibiotic subdirectories, "unnamed" is used
    for abx_names (length = 1)

    :param path: Path to directory containing antibiotic subdirectories
    :return: dict of abx_names: [image1_path, image2_path, etc.]
    """
    abx_names = [i for i in os.listdir(path) if not i.startswith('.') and os.path.isdir(os.path.join(path, i))]

    if not abx_names:
        abx_names = [""]

    plate_images_paths = {}
    for abx in abx_names:
        _path = os.path.join(path, abx)
        _temp_plate_images_paths = os.listdir(_path)
        _temp_plate_images_paths = [i for i in _temp_plate_images_paths if i.count('.jpg') > 0 or i.count('.JPG') > 0]
        _temp_plate_images_paths = [os.path.join(path, abx, i) for i in _temp_plate_images_paths]
        plate_images_paths[abx] = _temp_plate_images_paths

    return plate_images_paths



# Filename: 	model.py
# Author: 	Alessandro Gerada
# Date: 	2023-03-15
# Copyright: 	Alessandro Gerada 2023
# Email: 	alessandro.gerada@liverpool.ac.uk

"""Implementation of Model classes"""

from aigarmic._img_utils import convert_cv2_to_keras
import tensorflow as tf
import numpy as np
from typing import Optional
from abc import ABC, abstractmethod


class Model(ABC):
    def __init__(self, key: Optional[list[str]]):
        self.key = key

    def get_key(self) -> list[str]:
        """
        Return key to convert model output to human-readable label
        :return:
        """
        if not self.key:
            raise LookupError(
                """
                Unable to find an interpretation key to convert scores to label. Please provide one
                on Model class construction
                """
            )
        else:
            return self.key

    @abstractmethod
    def predict(self, image) -> dict:
        raise NotImplementedError


class KerasModel(Model):
    def __init__(self, path: str, trained_x: int, trained_y: int, key: Optional[list[str]]) -> None:
        """
        Base class for keras model to interpret colony image growth. Optionally provide a key to convert model output
        to a growth label, e.g.,
        ['No growth', 'Growth'] -> predictions of 0 will be interpreted as 'No growth', 1 as 'Growth'.
        ['No growth', 'Poor growth', 'Good growth'] -> predictions of 0 will be interpreted as 'No growth',
        1 as 'Poor growth', 2 as 'Good growth'.

        :param path: path to saved model
        :param trained_x: image width used to train model
        :param trained_y: image height used to train model
        :param key: key to interpret model output
        """
        self.path = path
        self.keras_data = tf.keras.models.load_model(path)
        if key:
            self.key = key
        else:
            # infer key from final output layer of loaded model
            final_layer = self.keras_data.layers[-1]
            inferred_key = [str(i) for i in range(final_layer.output.shape[1])]
            self.key = inferred_key
        self.trained_x = trained_x
        self.trained_y = trained_y
        super().__init__(key)

    def load_model(self, path: str) -> None:
        """
        Load a keras model from file

        :param path: path to saved model
        """
        self.keras_data = tf.keras.models.load_model(path)

    def predict(self, image: np.ndarray) -> dict:
        raise NotImplementedError


class SoftmaxModel(KerasModel):
    """
    SoftmaxModel is a one-stop model when more than one growth category is present, e.g.:
    ['No growth', 'Poor growth', 'Good growth']
    """

    def predict(self, image: np.ndarray) -> dict:
        """
        Predict growth category from image

        :param image: loaded using cv2.imread
        :return: dictionary with keys 'prediction', 'score', 'growth_code', 'growth', 'accuracy'
        """
        key = self.get_key()
        image = convert_cv2_to_keras(image, size_x=self.trained_x, size_y=self.trained_y)
        output = {'prediction': self.keras_data.predict(image)}
        output['score'] = tf.nn.softmax(output['prediction'])
        output['growth_code'] = np.argmax(output['score'])
        output['growth'] = key[output['growth_code']]
        output['accuracy'] = np.max(output['score'])
        return output


class BinaryModel(KerasModel):
    def __init__(self, path: str,
                 trained_x: int,
                 trained_y: int,
                 key: Optional[list[str]],
                 threshold: float = 0.5):
        """
        BinaryModel is a one-stop model when only two growth categories are present, e.g.:
        ['No growth', 'Growth'], or can be used in a two-step model:
        ['No growth', 'Growth'] -> ['Poor growth', 'Good growth']
        :param path: path to saved model
        :param trained_x: width of image used to train model
        :param trained_y: height of image used to train model
        :param key: key to interpret model output
        :param threshold: binary threshold to convert model output to growth code
        """
        if len(key) != 2:
            raise ValueError("Key of object of class BinaryModel must have length 2")
        if threshold < 0 or threshold > 1:
            raise ValueError("Binary threshold must be between 0 and 1")
        self.threshold = threshold
        super().__init__(path=path, key=key, trained_x=trained_x, trained_y=trained_y)

    def predict(self, image: np.ndarray) -> dict:
        """
        Predict growth category from image

        :param image: image loaded using cv2.imread
        :return: dictionary with keys 'prediction', 'score', 'growth_code', 'growth', 'accuracy'
        """
        key = self.get_key()
        image = convert_cv2_to_keras(image, size_x=self.trained_x, size_y=self.trained_y)
        prediction = self.keras_data.predict(image)
        [prediction] = prediction.reshape(-1)
        output = {'prediction': prediction}
        output['score'] = 0 if output['prediction'] <= self.threshold else 1
        output['growth_code'] = output['score']
        output['growth'] = key[output['growth_code']]
        output['accuracy'] = prediction if output['score'] == 1 else 1 - prediction
        return output


class BinaryNestedModel:
    def __init__(self, first_line_model: BinaryModel,
                 second_line_model: BinaryModel,
                 first_model_accuracy_acceptance: float = 0.9,
                 suppress_first_model_accuracy_check: bool = False) -> None:
        """
        Converts two BinaryModels into a two-step model. Generally this is used to support a more complex model for the
        second step of growth classification, which is generally more technically challenging. This approach also allows
        for computational efficiency by not running the second model in some situations (e.g., poor performance on the
        first model). Note that the key is inherited from the base models.

        :param first_line_model: BinaryModel object (generally differentiates growth from no growth)
        :param second_line_model: BinaryModel object (generally differentiates poor growth from good growth)
        :param first_model_accuracy_acceptance: minimum accuracy of first model to proceed to second model
        :param suppress_first_model_accuracy_check: if True, always proceed to second model
        """
        self.first_line_model = first_line_model
        self.second_line_model = second_line_model
        self.first_model_accuracy_acceptance = first_model_accuracy_acceptance
        self.suppress_first_model_accuracy_check = suppress_first_model_accuracy_check

        _key = self.second_line_model.get_key()
        _key.insert(0, self.first_line_model.get_key()[0])
        self.key = _key

    def predict(self, image: np.ndarray) -> dict:
        """
        Predict colony growth from image

        :param image: image loaded using cv2.imread
        :return: dictionary with keys 'prediction', 'score', 'growth_code', 'growth', 'accuracy'
        """
        first_line_classification = self.first_line_model.predict(image)
        if not self.suppress_first_model_accuracy_check and \
                first_line_classification['accuracy'] < self.first_model_accuracy_acceptance:
            # Do not try to make any additional predictions if first model poor
            return first_line_classification
        elif first_line_classification['growth_code'] == 0:
            return first_line_classification
        else:
            second_line_classification = self.second_line_model.predict(image)
            # Binary model assigns growth code 0 or 1, therefore need to 
            # correct second_line_classification['growth_code']
            second_line_classification['growth_code'] += 1
            second_line_classification['growth'] = self.second_line_model.get_key()[
                second_line_classification['growth_code']]
            return second_line_classification

    def get_key(self) -> list[str]:
        """
        Return key to convert model output to human-readable label
        
        :return: key (list of strings)
        """
        return self.key
